import typing
from .. import utils
from ..consts import Color, Flag

class Style(typing.TypedDict):
    height      : int    = None
    cursor      : str    = None
    show_help   : bool   = None
    cursor      : Color  = None
    symlink     : Color  = None
    directory   : Color  = None
    file        : Color  = None
    permissions : Color  = None
    selected    : Color  = None
    file_size   : Color  = None

def file(
    path      : str,
    all       : Flag[bool       ] = None,
    file      : Flag[bool       ] = None,
    directory : Flag[bool       ] = None,
    timeout   : Flag[int | float] = None,
    style     : Flag[Style      ] = None,
    __args__ : dict[str         ] = None
) -> list[str]:
    '''
    Prompt the user to select a file from the file tree.
    
    :param path: Path to begin traversing.
    :param all: Show hidden files.
    :param file: Allow files selection.
    :param directory: Allow directory selection.
    :param timeout: Timeout for the user to choose an option.
    :param style: Form style and color.
    
    :return: A list of pathes.
    '''
    
    process = utils.run(_file, __args__)
    
    return process.stdout.decode('utf-8').split('\n')


# Hack to differentiate arg and func
_file = file

# EOF


'''
    GUMX package
    
    https://github.com/egsagon/gumx
'''

import typing
from . import setup
from .consts import Error

# Setup GUM functions
if typing.TYPE_CHECKING:
    from .functions import *

# Run installation setup
setup.run()
__getattr__ = setup.arg_injector

# EOF


'''
    Function imports.
'''

from .choose  import choose
from .confirm import confirm
from .file    import file
from .search  import search
from .input   import input
from .pager   import pager
from .write   import write
from .spin    import spin

# EOF


import typing
from .. import utils
from ..consts import Color, Flag, Cursor

class Style(typing.TypedDict):
    width              : int    = None
    height             : int    = None
    show_cursor_line   : bool   = None
    show_line_numbers  : bool   = None
    show_help          : bool   = None
    base               : Color  = None
    header             : Color  = None
    placeholder        : Color  = None
    prompt             : Color  = None
    end_of_buffer      : Color  = None
    line_number        : Color  = None
    cursor_line_number : Color  = None
    cursor_line        : Color  = None
    cursor             : Cursor = None

def write(
    header      : Flag[str  ] = None,
    prompt      : Flag[str  ] = None,
    placeholder : Flag[str  ] = None,
    char_limit  : Flag[int  ] = None,
    value       : Flag[str  ] = None,
    style       : Flag[Style] = None,
    __args__    : dict[str  ] = None
) -> bool:
    '''
    Prompt for some multi-line text (ctrl+d to complete text entry).
    
    :param header: Header to print at the top of the form.
    :param placeholder: Text box placeholder.
    :param char_limit: Limit of characters to write.
    :param prompt: Prompt to display.
    :param value: Initial text box value.
    :param style: Form style and color.
    
    :return: The written content.
    '''
    
    output = utils.run(write, __args__).stdout.decode('utf-8')
    
    return output.strip()

# EOF


import typing
from .. import utils
from ..consts import Color, Flag

class Style(typing.TypedDict):
    affirmative : str    = None
    negative    : str    = None
    show_help   : bool   = None
    prompt      : Color  = None
    selected    : Color  = None
    unselected  : Color  = None

def confirm(
    message  : str,
    default  : Flag[str        ] = None,
    timeout  : Flag[int | float] = None,
    style    : Flag[Style      ] = None,
    __args__ : dict[str        ] = None
) -> bool:
    '''
    Confirm whether to perform an action.
    
    :param message: Message to display.
    :param default: Default confirmation action.
    :param timeout: Timeout for the user to choose an option.
    :param style: Form style and color.
    '''
    
    process = utils.run(confirm, __args__, throw = False)
    
    return process.returncode == 0

# EOF


import typing
from .. import utils
from ..consts import Color, Flag

class Style(typing.TypedDict):
    show_line_numbers : bool   = None
    help              : Color  = None
    line_number       : Color  = None
    match             : Color  = None
    match_highlight   : Color  = None

def pager(
    file      : str,
    soft_wrap : Flag[bool       ] = None,
    timeout   : Flag[int | float] = None,
    style     : Flag[Style      ] = None,
    __args__  : dict[str        ] = None
) -> None:
    '''
    Scroll through a long document.
    
    :param file: File to display.
    :param soft_wrap: Wether to soft wrap file lines.
    :param timeout: Timeout for the user to choose an option.
    :param style: Form style and color.
    '''
    
    # Not fun for large files, but file redirection does
    # not work the same on all platforms and i'm tired
    with open(file) as handle:
        __args__['file'] = handle.read()
    
    utils.run(pager, __args__, stdout = None)

# EOF


import typing
from .. import utils
from ..consts import Color, Flag, Cursor

class Style(typing.TypedDict):
    cursor      : Cursor = None
    width       : int    = None
    show_help   : bool   = None
    prompt      : Color  = None
    placeholder : Color  = None
    cursor      : Color  = None
    header      : Color  = None

def input(
    prompt      : Flag[str  ] = None,
    header      : Flag[str  ] = None,
    value       : Flag[str  ] = None,
    placeholder : Flag[str  ] = None,
    char_limit  : Flag[str  ] = None,
    password    : Flag[bool ] = None,
    style       : Flag[Style] = None,
    __args__    : dict[str  ] = None
) -> typing.Any:
    '''
    Prompt the user for input.
    
    :param prompt: Input field prompt.
    :param header: Input field header.
    :param value: Default input value.
    :param placeholder: Input field placeholder.
    :param char_limit: Max characters limit of the input.
    :param password: Wether to mask the input characters.
    :param style: Form style and color.
    
    :return: The written content.
    '''
    
    return utils.run(input, __args__).stdout.decode('utf-8').strip()

# EOF


import typing
from .. import utils
from ..consts import Color, Flag

class Style(typing.TypedDict):
    height            : int   = None
    cursor_prefix     : str   = None
    selected_prefix   : str   = None
    unselected_prexix : str   = None
    show_help         : bool  = None
    item              : Color = None
    cursor            : Color = None
    header            : Color = None
    selected          : Color = None

def choose(
    items    : typing.Iterable,
    header   : Flag[str            ] = None,
    limit    : Flag[int            ] = None,
    ordered  : Flag[bool           ] = None,
    selected : Flag[typing.Iterable] = None,
    timeout  : Flag[int | float    ] = None,
    style    : Flag[Style          ] = None,
    __args__ : dict[str            ] = None
) -> list:
    '''
    Choose an option from a list of choices.
    
    :param items: Iterable of items to choose from.
    :param header: Header to print at the top of the form.
    :param limit: Multiple selection limit. 0 = No limit.
    :param ordered: Wether to keep the order of selected options.
    :param selected: Item(s) selected by default.
    :param timeout: Timeout until for the user to choose an option.
    :param style: Form style and color.
    
    :return: List of selected items.
    '''
    
    raw_items = __args__['items'] = utils.dump(items)
    
    choices = utils.run(choose, __args__).stdout.decode('utf-8').split('\n')
    
    return [i for i, r in zip(items, raw_items) if r in choices]

# EOF


'''
    GUMX setup.
'''

import os
import glob
import typing
import inspect
import zipfile

from . import utils
from . import consts
from . import functions


def download() -> str:
    '''
    Download the latest version of the GUM archive to
    the GUM user data dir.
    Returns the extracted folder location.
    '''
    
    print('[GUMX] Fetching latest GUM executable')
    assets = utils.get(consts.API).json()[0]['assets']
    
    # Fetch versions
    versions = list(filter(lambda asset: (
        asset['name'].endswith(('tar.gz', 'zip'))
        and consts.OS in asset['name']
        and consts.ARCH in asset['name']
    ), assets))
    
    assert len(versions), 'Could not find a compatible version'
    version = versions[0] # Most recent version
        
    # Download archive
    archive_path = os.path.join(consts.DIR, version['name'])
    install_path = os.path.join(consts.DIR, 'gum')
    
    print('[GUMX] Downloading release', version['name'])
    utils.download(version['browser_download_url'], archive_path)
    
    # Extract archive
    with zipfile.ZipFile(archive_path) as archive:
        archive.extractall(install_path)
    
    # Cleanup
    os.remove(archive_path)
    return install_path

def find(location: str) -> str:
    '''
    Find GUM executable in a folder.
    '''
    
    pattern = os.path.join(location, '**/gum*')
    
    _, *files = glob.glob(pattern, recursive = True)
    
    return files[0]

def run() -> None:
    '''
    Setup the GUM executable.
    '''
    
    install_link = os.path.join(consts.DIR, 'executable')
    
    if not os.path.exists(install_link):
        consts.INSTALL = find(download())
        print('[GUMX] Installation complete')
        
        with open(install_link, 'w') as file:
            file.write(consts.INSTALL)
            
            print(consts.INSTALL)
    
    else:
        with open(install_link) as file:
            consts.INSTALL = file.read()

def arg_injector(name: str) -> typing.Callable:
    '''
    Append an `__args__` argument on function calls.
    '''
    
    def wrapper(*args, **kwargs) -> None:
        func: typing.Callable = getattr(functions, name)
        arguments = inspect.signature(func).bind(*args, **kwargs).arguments
        
        return func(**(arguments | {'__args__': arguments}))
    
    return wrapper

# EOF


'''
    GUMX constants.
'''

import os
import typing
import platform
import platformdirs


# Installation constants
API = 'https://api.github.com/repos/charmbracelet/gum/releases'

OS = platform.system()

ARCH = {'amd64': 'x86_64', 'x86': 'i386', 'aarch64': 'arm64'} \
    [platform.machine().lower()]

DIR = platformdirs.user_data_dir('gumx')
os.makedirs(DIR, exist_ok = True)

INSTALL: str = None


# Type definitions
Flag = typing.Optional

class Error(Exception):
    '''
    An error raised by GUM.
    '''

class Color(typing.TypedDict):
    foreground = None
    background = None

class Cursor(Color):
    mode: typing.Literal['blink'] = None

# EOF


import typing
import subprocess
import contextlib
from .. import utils
from .. import consts
from ..consts import Color, Flag

class Style(typing.TypedDict):
    spinner : Color = None
    title   : Color = None

@contextlib.contextmanager
def spin(
    title    : Flag[str                   ] = None,
    spinner  : Flag[typing.Literal['dot'] ] = None,
    align    : Flag[typing.Literal['left']] = None,
    timeout  : Flag[int | float           ] = None,
    style    : Flag[Style                 ] = None,
    __args__ : dict[str                   ] = None
):
    '''
    Display spinner while running a script.
    
    :param title: Text to display while spining.
    :param spinner: Spinner type.
    :param align: Spinner alignment.
    :param timeout: Timeout until abort.
    :param style: widget type.
    '''
    
    try:
        cmd = [consts.INSTALL, 'spin'] + utils.dump_args(spin, __args__) + ['--']
        
        process = subprocess.Popen()
        
        yield
    
    finally:
        pass

# EOF


import typing
from .. import utils
from ..consts import Color, Flag

class Style(typing.TypedDict):
    indicator          : Color  = None
    selected_indicator : Color  = None
    unselected_prefix  : Color  = None
    header             : Color  = None
    text               : Color  = None
    cursor_text        : Color  = None
    match              : Color  = None
    prompt             : Color  = None
    placeholder        : Color  = None

def search(
    items       : typing.Iterable,
    prompt      : Flag[str        ] = None,
    header      : Flag[str        ] = None,
    placeholder : Flag[str        ] = None,
    value       : Flag[str        ] = None,
    reverse     : Flag[bool       ] = None,
    fuzzy       : Flag[bool       ] = None,
    sort        : Flag[bool       ] = None,
    timeout     : Flag[int | float] = None,
    style       : Flag[Style      ] = None,
    __args__ : dict[str           ] = None
) -> typing.Any:
    '''
    Search in a list of values with fuzzy matching.
    
    :param items: Items to filter.
    :param header: Header to print at the top of the form.
    :param placeholder: Form placeholder.
    :param prompt: Text to display on top of the form.
    :param value: Initial filter value.
    :param reverse: Wether to reverse filtering order.
    :param fuzzy: Wether to allow fuzzy sorting.
    :param sort: Wether to sort the results.
    :param timeout: Timeout for the user to interact.
    :param style: Form style and color.
    
    :return: List of filtered items.
    '''
    
    raw_items = __args__['items'] = utils.dump(items)
    
    output = utils.run(search, __args__).stdout.decode('utf-8').strip()
    
    for i, r in zip(items, raw_items):
        if r == output:
            return i

# EOF


import typing
import requests
import subprocess

from . import consts

def get(url: str, **kw) -> requests.Response:
    '''
    GET request wrapper.
    '''
    
    response = requests.get(url, **kw)
    response.raise_for_status()
    return response

def download(url: str, path: str, chunk_size = 8192) -> None:
    '''
    Download a large file.
    '''
    
    with get(url, stream = True) as stream, open(path, 'wb') as file:
        for chunk in stream.iter_content(chunk_size):
            file.write(chunk)

def dump(obj: typing.Any) -> str | list[str]:
    '''
    Dump a single value into string.
    
    If the object is an iterable, objects inside will be dumped.
    '''
    
    if obj is True:
        return ''
    
    if isinstance(obj, str):
        return obj
    
    if isinstance(obj, dict):
        obj = [f'{k} - {v}' for k, v in obj.items()]
    
    if isinstance(obj, tuple | list):
        return list(map(dump, obj))
    
    return repr(obj)

def flatten(obj: dict, t: dict = None, r: str = '') -> dict:
    '''
    Flatten a style dict.
    '''
    
    track = t or {}
    
    for k in obj:
        if isinstance(obj[k], dict):
            flatten(obj[k], track, r + k + '.')
        
        else:
            track[r + k] = obj[k]
    
    return track

def dump_args(func: typing.Callable, args: dict[str, typing.Any]) -> list[str]:
    '''
    Dump arguments into a shell command.
    '''
    
    t = func.__annotations__
    
    # Construct args chain
    output = []
    
    for arg, value in args.items():
        if arg == 'style' or value in (None, False): continue
        
        raw = dump(value)
        
        if 'Optional' in str(t[arg]):
            output.append(f'--{arg}')
        
        if isinstance(raw, list):
            output += raw
        elif raw != '':
            output.append(raw)
    
    for arg, value in flatten(args.get('style', {})).items():
        output.append(f'--{arg}')
        output.append(dump(value))
    
    return output

def run(
    func: typing.Callable,
    args: dict[str, typing.Any],
    throw: bool = True,
    **proc_kwargs
) -> subprocess.CompletedProcess:
    '''
    Run a GUM command.
    '''
    
    cmd = [consts.INSTALL, func.__name__] + dump_args(func, args)
    
    print('[EXEC]', cmd)
    
    process = subprocess.run(**({
        'args': cmd,
        'stdout': subprocess.PIPE
    } | proc_kwargs))
    
    if throw and process.returncode != 0:
        raise consts.Error(f'Error code {process.returncode}')
    
    return process

# EOF